/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code.
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *		Karl Guertin <grayrest@grayrest.com>
 *		James A. Overton <james@overton.ca>
 *		Christian Stocker (Bitflux)
 *		Paul Everitt
 *		David Palm
 *
 * ***** END LICENSE BLOCK ***** */

/**
@fileoverview

Mozile Parse Mozile Editing Scheme Javascript

Created by James A. Overton
Created on 2004-03-24

Last revised by James A. Overton
Last revised on 2004-11-16

Last revised by Richard Prescott
Last revised on 2004-04-18
Added some namespace support

HISTORY:
	Some code in this file was first found in mozilehtmltb.js, while the 
majority was cut out of mozileTB.js and separated into different files.

DESCRIPTION:
	This code is responsible for parsing the special Mozile Editing Scheme XML
files. There's nothing very fancy going on here, but it is a lot of work.

*/

/**
Create MES data structure.

@constructor
*/
function MozileMES( images, commands, tags, toolbar, eDOM ) {
	this.images = images || new Array();
	this.commands = commands || new Array();
	this.tags = tags || new Array();
	this.toolbar = toolbar || new Array();
	this.eDOM = eDOM || new Object();
}

//Global methods
var mozileMES = new MozileMES();
mozileMES.loadMES = mozileLoadMES;
mozileMES.findByName = mozileFindByName;

/**
Load and parse the MES configuration file.
Amongst other things this defines the commands and toolbar buttons.

<p>The toolbar data is stored in the target document:<br />
<code>mozileInjectTarget.document.mozileMES</code>

@param mozileInjectTarget target window being edited.
@param mes MES document to parse.
*/
function mozileLoadMES(mozileInjectTarget, mes) {
	var f = ["mozileParseMES.js","mozileLoadMES"];

	var MES, MES_doc, mesloaded;

	var mozileIT=mozileInjectTarget;
	
	if ( mes=="XHTML" || !mes || mes=="" ) {
		mozileDebug(f,3,"Loading XHTML");
		MES_doc = MOZILEC.MES_DEFAULT;
	} else {
		MES_doc = mes;
	}
	mozileDebug(f,3,"MES to load: "+ MES_doc);
	
	mozileInjectTarget.document.mozileMES = new MozileMES();
	var mozileMESi = mozileInjectTarget.document.mozileMES;

	mesloaded = false;
	try {
  	MES = mozileInjectTarget.document.implementation.createDocument("", "", null);
		MES.async = false;
		mesloaded = MES.load(MES_doc);
		if(!mesloaded) {
			mozileDebug(f,1,"Error loading MES.");
			MES=null;		
		}
	}
	catch(e)
	{
		mozileDebug(f,1,"Exception loading MES: "+e);
		MES = null;
	}

	//If error or XML parsing error...
	if ( MES==null || !MES.documentElement || MES.documentElement.tagName == "parsererror") {
		mozileDebug(f,2,"Could not load MES");
		if ( MES_doc !== MOZILEC.MES_DEFAULT ) {
			mozileLoadMES(mozileIT,"XHTML");  //attempt to load the default MES.
		}
		return;
	}
	else {	// success!	
		try {
			mozileMESi.images = mozileParseIcons(MES);
			mozileMESi.commands = mozileParseCommands(MES);
			mozileMESi.tags = mozileParseTags(MES);
			mozileMESi.toolbar = mozileParseToolbar(MES);
			mozileMESi.eDOM = mozileParseEDOM(MES);

			mozileParseSpecialChars(mozileInjectTarget, MES);

			mozileDebug(f,3,"MES update completed.");
		} catch(e) {
			mozileDebug(f,1,"Exception updating MES: "+e);
		}
	}
}
/**
Parse fixed toolbar command elements.
Looks for fixed toolbar items to be disabled.

<p>Example command element format:<br />
 &lt;command name="bug" disabled="true" &gt;

@return array of names as strings.
@type Array
*/
function mozileParseToolbar(MES) {
	var f = ["mozileParseMES.js","mozileParseToolbar"];
	var id = new Array();
 	try{
		var commands = MES.documentElement.getElementsByTagName("commands");
		if (commands.length > 0) {
			var command = commands[0].getElementsByTagName("command");
			for( var i=0; i<command.length; i++ ) {
				if ( command[i].getAttribute("type") == "toolbar" ) {
					var name = command[i].getAttribute("name");
					var disabled = command[i].getAttribute("disabled");
					if ( name && disabled=="true" ) {
						if (mozileGetToolbarById("mozile-"+name)) {
							id.push(name);
						}
					}
				}
			}
		}
	}catch(e){
		mozileDebug(f,1,"Exception parsing MES fixed toolbar command: "+e);
	}
	return id;
}

/**
Parse eDOM part of the MES.
Not strict parsing or validation of the xhtml value, so get it right...

@return xhtml default container.
@type Object.
*/
function mozileParseEDOM(MES) {
	var f = ["mozileParseMES.js","mozileParseEDOM"];
	
	var mesEDOM = new Object();
	mesEDOM.xhtmlDefaultContainer = "div";
	try {
		var dc = MES.getElementsByTagName("container");
		if (dc.length>0 && dc[0].hasAttribute("default")) {
			var c = dc[0].getAttribute("default");
			//Could check is a valid xhtml element.
			mesEDOM.xhtmlDefaultContainer = c;
		}
	} catch(e) {
		mozileDebug(f,1,"Exception parsing eDOM MES: "+e);
	}
	return mesEDOM;
}

function mozileParseIcons(MES) {
	var icons = MES.documentElement.getElementsByTagName("icons")[0];
	var images = new Array();
	images[0] = icons.getAttribute("small");
	images[1] = icons.getAttribute("large");
	images[2] = icons.getAttribute("format");
	return images;
}

function mozileParseCommands(MES) {
	var commands = MES.documentElement.getElementsByTagName("commands")[0];
	var commandTags = commands.childNodes;
	var commandTag, parsed, i;
	var commandArray = new Array();

	for(i=0; i < commandTags.length; i++) {
		commandTag=commandTags[i];
		if( (commandTag.nodeType == Node.ELEMENT_NODE) && (commandTag.tagName == "command") ) {
			parsed = parseCommand(commandTag);
			if(parsed) {
				commandArray.push( parsed );
			}
		}
	}

 //This is a private function of mozileParseCommands.
 function parseCommand(tag) {
 	var f = ["mozileParseMES.js","parseCommand"];
	var carr = new Array();
	var type = tag.getAttribute("type");
		
 	// If fixed toolbar item, ignore it.
	if (type == "toolbar")
		return null;

		// deal with common attributes
	carr[0] = tag.getAttribute("name");
	carr[1] = tag.getAttribute("title");
	carr[2] = tag.getAttribute("tooltip");
	carr[3] = tag.getAttribute("icon");
	carr[4] = tag.getAttribute("modifiers");
	carr[5] = tag.getAttribute("key");
	carr[6] = tag.getAttribute("type");
	
		// deal with special attributes
	var args=new Array();

	switch(type)
	{
		// style is something specific to XHTML
		// no namespace is involved
		case "toggleStyle":
			args[0] = tag.getAttribute("property");
			args[1] = tag.getAttribute("value");
			args[2] = tag.getAttribute("default");
			args[3] = tag.getAttribute("activeproperty");
			//
			var re = /\s*,\s*/;
			args[0] = args[0].split(re,10);
			args[1] = args[1].split(re,10);
			// default value may be optional at some time in the future...
			if ( args[2] && args[2]!="" ) {
				args[2] = args[2].split(re,10);
			}else{
				args[2]=new Array();
			}
			//Number of properties used when updating active toolbar.
			if ( !args[3] ) {
				args[3] = args[0].length;
			}
			break;
			
		// XXX: toggleAttribute doesn't care about attribute's namespace
		// it is not a breaking issue for now  - Richard
		case "toggleAttribute":
			args[0] = tag.getAttribute("attribute");
			args[1] = tag.getAttribute("value");
			args[2] = tag.getAttribute("default");
			break;

		// style is something specific to XHTML
		// no namespace is involved
		case "styleText":
			args[0] = tag.getAttribute("property");
			args[1] = tag.getAttribute("value");
			break;

		// style is something specific to XHTML
		// no namespace is involved
		case "styleLines":
			args[0] = tag.getAttribute("property");
			args[1] = tag.getAttribute("value");
			break;

		case "setLinesContainer":
			args[0] = tag.getAttribute("namespace");
			args[1] = tag.getAttribute("tag");
			break;

		case "wrapText":
			args[0] = tag.getAttribute("namespace");
			args[1] = tag.getAttribute("tag");
			break;

		// XHTML ul and ol only. 
		case "toggleListLines":
			args[0] = tag.getAttribute("tag");
			args[1] = tag.getAttribute("old");
			break;

		case "indent":
		case "outdent":
			args[0] = tag.getAttribute("value");
			break;

		case "indentLists":
		case "outdentLists":
			break;

		case 'link':
			break;

		case "unlink":
			break;

		case "unformat":
			break;

		case 'table':
			break;

		case 'image':
			break;

		case 'hr':
			break;

		case 'x-element':
			args[0] = tag.getAttribute("tag");
			break;
			
		case 'mozileCommand':
			args[0] = tag.getAttribute("commandname");
			break;

			// if the command type hasn't been found, throw an error
		default:
			mozileDebug(f,1,'Command type "'+ type +'" not defined!');
			return null;
	}
	
	carr[7] = args;
		// done
	return carr;
 }

 return commandArray;
} //End of private functions.

function mozileParseTags(MES) {
	var tags = MES.documentElement.getElementsByTagName("tags")[0];
	var tagTags = tags.childNodes;
	var parsed, tagTag, i, j;
	var tagArray = new Array();
	
	for(j=0; j < tagTags.length; j++) {
		tagTag=tagTags[j];
		if( (tagTag.nodeType == Node.ELEMENT_NODE) && (tagTag.tagName == "tag") ) {
			parsed = parseTag(tagTag, tagArray)
			if(parsed) {
				tagArray.push( parsed );
			}
		}
	}
// NB : functions below are private to mozileParseTags.

	// this is trickier, because of the "use" attribute
	// keys/commands in the array already are overridden,
	// those not in the array are tacked onto the end...
	// this could be improved, but I'm lazy.
function parseTag(tag, tagArray) {
	var tarr = new Array();
	var use, useIndex;
		
		// deal with common attributes
	tarr[0] = tag.getAttribute("name");
	tarr[1] = tag.getAttribute("title");
	tarr[2] = tag.getAttribute("unique");
	
	//alert("Parsing Tag Tag <"+tarr[0]+">");
		// check for duplicates
	if(mozileMES.findByName(tarr[0],tagArray)>=0) {
		alert("Tag "+ tarr[0] +" is defined more than once!");
		return 0;
	}
	
		// if we are copying a whole tag, then find it and copy it, then return
	use = tag.getAttribute("use");
	if(use && use!="") {
		useIndex=mozileMES.findByName(use, tagArray);
		//alert("Reuse all from "+ use + " at " +useIndex);
		if(useIndex >= 0) {
			tarr[3]=mozileArrayCopy(tagArray[useIndex][3]);
			tarr[4]=mozileArrayCopy(tagArray[useIndex][4]);
			return tarr;
		} else {
			alert("The "+tarr[0]+" tag tells me to 'use' the commands and keys from "+use+" tag, but they are not yet defined!");
		}
	}
	
		// get the "commands" and "keys" tags
	var children = tag.childNodes;
	var commandTags = new Array();
	var commands = new Array();
	var reuseCommands = 0;
	var keyTags = new Array();
	var keys = new Array();
	var reuseKeys = 0;
	for(var i=0; i<children.length; i++) {
		if( (children[i].nodeType == Node.ELEMENT_NODE) && (children[i].tagName == "commands") ) {
			commandTags = children[i].childNodes;
			use = children[i].getAttribute("use");
			if(use && use!="") {
				useIndex=mozileMES.findByName(use,tagArray);		// get the index of the tag name
				if(useIndex >= 0) {
					commands=mozileArrayCopy(tagArray[useIndex][3]);	// the third entry holds commands
					reuseCommands=1;
				} else {
					alert("The "+tarr[0]+" tag tells me to 'use' the commands from "+use+" tag, but they are not yet defined!");
				}
			}
		}
		if( (children[i].nodeType == Node.ELEMENT_NODE) && (children[i].tagName == "keys") ) {
			keyTags = children[i].childNodes;
			use = children[i].getAttribute("use");
			if(use && use!="") {
				useIndex=mozileMES.findByName(use,tagArray);		// get the index of the tag name
				if(useIndex >= 0) {
					keys=mozileArrayCopy(tagArray[useIndex][4]);	    // the fourth entry holds keys
					reuseKeys=1;
				} else {
					alert("The "+tarr[0]+" tag tells me to 'use' the keys from "+use+" tag, but they are not yet defined!");
				}
			}
		}
	}
	
	//alert("Reuse? "+reuseCommands+reuseKeys);
		// parse all the command tags into the commands array
	for(var j=0; j<commandTags.length; j++) {
		tag = commandTags[j];
		if ( tag.nodeType == Node.ELEMENT_NODE ) {
  		if(reuseCommands) {
  			if(tag.tagName=="command") {
  				commandIndex = mozileMES.findByName(tag.getAttribute("name"), commands);
  				if(commandIndex >= 0) {
  					commands[commandIndex] = parseTagCommand(tag);
  				} else {
  					commands.push( parseTagCommand(tag) );
  				}
  			}
  			if(tag.tagName=="commandGroup") {
  				commandIndex = mozileMES.findByName(tag.getAttribute("name"), commands);
  				if(commandIndex >= 0) {
  					commands[commandIndex] = parseTagCommandGroup(tag);
  				} else {
  					commands.push( parseTagCommandGroup(tag) );
  				}
  			}
  		}
  		else {
  			if(tag.tagName=="command") {
  				commands.push( parseTagCommand(tag) );
  			}
  			if(tag.tagName=="commandGroup") {
  				commands.push( parseTagCommandGroup(tag) );
  			}
  		}
		}
	}

	tarr[3] = commands;
	
		// parse all the key tags into the keys array
	for(var k=0; k<keyTags.length; k++) {
		tag = keyTags[k];
		if ( tag.nodeType == Node.ELEMENT_NODE ) {
  		if(reuseKeys) {
  			if(tag.tagName=="key") {
  				//alert("Reuse KeyTag "+tag.getAttribute("name"));
  				keyIndex = findByName(tag.getAttribute("name"), keys);
  				if(keyIndex >= 0) {
  					//alert("Replace key at "+keyIndex);
  					keys[keyIndex] = parseTagKey(tag);
  				} else {
  					keys.push( parseTagKey(tag) );
  				}
  			}
  		} else {
  			if (tag.tagName=="key") {
  				//alert("New KeyTag "+tag.getAttribute("name"));
  				keys.push( parseTagKey(tag) );
  			}
  		}
		}
	}
	
	tarr[4] = keys;
	
		// done
	return tarr;
	
}

function parseTagCommand(tag) {
	var carr = new Array();
	carr[0] = tag.getAttribute("name");
	carr[1] = tag.getAttribute("accesskey");
	carr[2] = tag.getAttribute("button");
	return carr;
}

function parseTagCommandGroup(tag) {
	var garr = new Array();
	garr[0] = tag.getAttribute("name");
	garr[1] = tag.getAttribute("title");
	garr[2] = tag.getAttribute("tooltip");
	garr[3] = tag.getAttribute("icon");
	garr[4] = tag.getAttribute("accesskey");
	garr[5] = tag.getAttribute("button");

	var commands = new Array();
	var children = tag.childNodes;
	var i;
	for(i=0; i<children.length; i++) {
		if( (children[i].nodeType == Node.ELEMENT_NODE) && (children[i].tagName == "command") ) {
			commands.push( parseTagCommand(children[i]) );
		}
	}

	garr[6] = commands;
	
	return garr;
}

function parseTagKey(tag) {
	var karr = new Array();
	karr[0] = tag.getAttribute("name");
	karr[1] = tag.getAttribute("command");
	return karr;
}

	return tagArray;
} //end private functions


/**
Find the named command in an array, return the index.
*/
function mozileFindByName(name, array) {
	for(var i=0; i<array.length; i++) {
		if(array[i][0]==name) {
			return i;
		}
	}
		// nothing found
	return -1;
}

/**
Simple array copy function.
Won't copy a recursive array.

@param {array} a array to be copied.
@return one level deep copy of array.
@type array
*/
function mozileArrayCopy(a) {
  var a2 = new Array();
	return a2.concat(a);
}
